use std::{
	convert::TryInto,
	ffi::{
		CString,
		c_void,
	},
	time::Duration,
};

use anyhow::{
	anyhow,
	Context,
};
use sdl2::{
	event::Event,
	keyboard::{Keycode, Scancode},
};

use opengl_rust::{
	glezz,
	gpu_buffers,
	shader,
	timestep::TimeStep,
};

fn main () -> anyhow::Result <()> {
	let sdl_context = sdl2::init ().map_err (|e| anyhow! ("Can't init SDL: {}", e))?;
	
	let video_subsystem = sdl_context.video ().map_err (|e| anyhow! ("Can't get SDL video subsystem: {}", e))?;
	let window = video_subsystem.window ("Heightmap terrain demo", 960, 540)
		.position_centered ()
		.opengl ()
		.build ()
	?;
	
	gl::load_with (|s| {
		video_subsystem.gl_get_proc_address (s) as *const _
	});
	
	assert! (gl::ClearColor::is_loaded ());
	
	let gl_ctx = window.gl_create_context ().map_err (|e| anyhow! ("Can't create OpenGL context: {}", e))?;
	
	window.gl_make_current (&gl_ctx).map_err (|e| anyhow! ("Can't make OpenGL context current: {}", e))?;
	
	let mut time_step = TimeStep::new (60, 1000);
	
	let mut graphics_frames = 0;
	
	let controller_subsystem = sdl_context.game_controller ().unwrap ();
	let controller = controller_subsystem.open (0).ok ();
	
	let mut event_pump = sdl_context.event_pump ().unwrap ();
	
	let vertex_shader = shader::ShaderObject::from_source (gl::VERTEX_SHADER, b"
#define lowp
#define mediump
#define highp
#line 0
attribute highp vec4 attr_pos;

void main (void) {
	gl_Position = attr_pos;
}
	").map_err (|e| anyhow! ("Can't create vertex shader: {}", e))?;
	
	let fragment_shader = shader::ShaderObject::from_source (gl::FRAGMENT_SHADER, b"
#define lowp
#define mediump
#define highp
#line 0
void main (void) {
	gl_FragColor = vec4 (0.329, 0.431, 0.478, 1.0);
}
	").map_err (|e| anyhow! ("Can't create fragment shader: {}", e))?;
	
	let shader_program = shader::ShaderProgram::new (&vertex_shader, &fragment_shader).map_err (|e| anyhow! ("Can't link shader program: {}", e))?;
	
	let attr_pos = shader_program.get_attribute_location (&CString::new ("attr_pos".as_bytes ())?);
	
	let vertex_buffer = gpu_buffers::VertexBuffer::from_slice (&[
		0.0, 0.0,
		1.0, 0.0,
		1.0, 1.0,
		0.0, 1.0,
	]);
	
	let index_buffer = gpu_buffers::IndexBuffer::from_slice_u32 (&[
		0, 1, 2,
		0, 2, 3
	]);
	
	'running: loop {
		let frames_to_do = time_step.step ();
		
		let _mouse = event_pump.mouse_state ();
		
		for event in event_pump.poll_iter() {
			match event {
				Event::Quit {..} |
				Event::KeyDown { keycode: Some (Keycode::Escape), .. } => {
					break 'running
				},
				_ => (),
			}
		}
		
		for _ in 0..frames_to_do {
			
		}
		
		window.gl_make_current (&gl_ctx).unwrap ();
		glezz::clear_color (0.392f32, 0.710f32, 0.965f32, 1.0f32);
		glezz::clear (gl::COLOR_BUFFER_BIT | gl::DEPTH_BUFFER_BIT | gl::STENCIL_BUFFER_BIT);
		
		shader_program.use_program ();
		glezz::enable_vertex_attrib_array (Some (attr_pos.try_into ()?));
		
		vertex_buffer.bind ();
		index_buffer.bind ();
		
		unsafe {
			gl::VertexAttribPointer (attr_pos.try_into ()?, 2, gl::FLOAT, 0, 4 * 2, 0 as *const u8 as *const c_void);
			
			gl::DrawRangeElements (gl::TRIANGLES, 0, 6, 6, gl::UNSIGNED_INT, 0 as *const u8 as *const c_void);
		}
		
		window.gl_swap_window ();
		graphics_frames += 1;
		
		std::thread::sleep (Duration::from_millis (15));
	}
	
	Ok (())
}
